package edu.jhmi.telometer.calc.impl;

import edu.jhmi.telometer.bean.Dimension;
import edu.jhmi.telometer.bean.Image;
import edu.jhmi.telometer.bean.Mask;
import edu.jhmi.telometer.bean.Pixelation;
import edu.jhmi.telometer.bean.Point;
import edu.jhmi.telometer.bean.Rect;
import edu.jhmi.telometer.calc.api.MaskConvertUtil;
import edu.jhmi.telometer.calc.api.PositionConvertUtil;
import edu.jhmi.telometer.calc.api.MaskLogicUtil;
import ij.ImagePlus;
import ij.process.ByteProcessor;
import ij.process.ImageProcessor;
import org.springframework.util.Assert;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class MaskConvertUtilImpl implements MaskConvertUtil {


    private PositionConvertUtil positionConvertUtil = new PositionConvertUtilImpl();
    private MaskLogicUtil maskLogicUtil = new MaskLogicUtilImpl();


    @Override
    public Set<Point> maskToPoints(Mask mask) {
        Set<Point> points = new HashSet();
        List<Boolean> bools = mask.getBools();
        for (int position = 0; position < bools.size(); position++){
            if (bools.get(position)){
                Point point = positionConvertUtil.positionToPoint(position, mask.getDimension());
                points.add(point);
            }
        }
        return points;
    }

    @Override
    public Mask pointsToMask(Set<Point> points, Dimension dimension) {
        Mask mask = maskLogicUtil.getEmptyMask(dimension);
        List<Boolean> bools = mask.getBools();
        for (Point point : points){
            int position = positionConvertUtil.pointToPosition(point, dimension);
            bools.set(position, true);
        }
        return Mask.builder().dimension(dimension).bools(bools).build();
    }



    @Override
    public Pixelation imageAndMaskToPixelation(Image image, Mask mask) {
        List<Integer> pixels = new ArrayList();
        int size = mask.getBools().size();
        for (int i = 0; i < size; i++){
            boolean inMask = mask.isPositionInMask(i);
            if (inMask){
                int originalValue = image.getPixelation().getPixels().get(i);
                pixels.add(originalValue);
            }
        }
        Pixelation pixelation = Pixelation.builder().pixels(pixels).build();
        return pixelation;
    }

    @Override
    public ImagePlus maskToImagePlus(Mask mask) {
        byte[] bytes = maskToBytes(mask);
        Dimension dimension = mask.getDimension();
        ImagePlus imagePlus = new ImagePlus("maskToImagePlus", new ByteProcessor(dimension.getWidth(), dimension.getHeight(), bytes));
        return imagePlus;
    }


    private byte[] maskToBytes(Mask mask) {
        final byte ON = -1; //WHITE
        final byte OFF = 0; //BLACK
        int size = mask.getBools().size();
        byte[] bytes = new byte[size];
        for (int i = 0; i < size; i++){
            bytes[i] = mask.getBools().get(i) ? ON : OFF;
        }
        return bytes;
    }

    @Override
    public Mask byteImageProcessorToMask(ImageProcessor imageProcessor) {
        byte[] bytes = (byte[]) imageProcessor.getPixels();
        List<Boolean> bools = new ArrayList<>();
        for (byte b : bytes){
            boolean value =  (b == 0) ? false : true;
            bools.add(value);
        }
        Dimension dimension = Dimension.builder().width(imageProcessor.getWidth()).height(imageProcessor.getHeight()).build();
        Mask mask = Mask.builder().bools(bools).dimension(dimension).build();
        return mask;
    }

    @Override
    public Mask nonZeroImageToMask(Image image) {
        List<Boolean> booleans = new ArrayList<>();
        Pixelation pixelation = image.getPixelation();
        for (int pixel : pixelation.getPixels()){
            boolean value =  (pixel != 0);
            booleans.add(value);
        }
        Mask mask = Mask.builder().bools(booleans).dimension(image.getDimension()).build();
        return mask;
    }

    @Override
    public Mask rectToRectPartialMask(Rect rect) {
        return maskLogicUtil.getFullMask(positionConvertUtil.rectToDimension(rect));
    }

    @Override
    public Mask rectPartialMaskToFullMask(Mask partialRectMask, Rect rect, Image image) {
        Assert.isTrue(partialRectMask.getBools().size() == (rect.getHeight() * rect.getWidth()),
                "rectMask length is equal to rect width*height");
        Mask mask = maskLogicUtil.getEmptyMask(image.getDimension());
        for (int rectangleY = 0; rectangleY < partialRectMask.getDimension().getHeight(); rectangleY++) {
            for (int rectangleX = 0; rectangleX < partialRectMask.getDimension().getWidth(); rectangleX++) {
                Point relativePoint = Point.builder().x(rectangleX).y(rectangleY).build();
                Point absolutePoint = positionConvertUtil.rectPointToAbsolutePoint(rect, relativePoint);
                int absolutePosition = positionConvertUtil.pointToPosition(absolutePoint, image.getDimension());
                int relativePosition = positionConvertUtil.pointToPosition(relativePoint, positionConvertUtil.rectToDimension(rect));
                boolean assignedValue = partialRectMask.getBools().get(relativePosition);
                mask.getBools().set(absolutePosition, assignedValue);
            }
        }
        return mask;
    }

    @Override
    public Point averagePoint(Mask mask) {
        Set<Point> points = maskToPoints(mask);
        int totalX = 0;
        int totalY = 0;
        for (Point point : points){
            totalX += point.getX();
            totalY += point.getY();
        }
        int averageX = (totalX / points.size());
        int averageY = (totalY / points.size());
        return Point.builder().x(averageX).y(averageY).build();
    }

}
