package edu.jhmi.telometer.service.impl;

import edu.jhmi.telometer.bean.Cell;
import edu.jhmi.telometer.bean.CellType;
import edu.jhmi.telometer.bean.LesionType;
import edu.jhmi.telometer.bean.AdminType;
import edu.jhmi.telometer.bean.Preset;
import edu.jhmi.telometer.bean.Project;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.bean.Telomere;
import edu.jhmi.telometer.bean.TissueType;
import edu.jhmi.telometer.interfce.Named;
import edu.jhmi.telometer.repo.CellRepo;
import edu.jhmi.telometer.repo.CellTypeRepo;
import edu.jhmi.telometer.repo.LesionTypeRepo;
import edu.jhmi.telometer.repo.PresetRepo;
import edu.jhmi.telometer.repo.ProjectRepo;
import edu.jhmi.telometer.repo.ScoringRepo;
import edu.jhmi.telometer.repo.TelomereRepo;
import edu.jhmi.telometer.repo.TissueTypeRepo;
import edu.jhmi.telometer.service.api.AdminService;
import edu.jhmi.telometer.util.api.ImageDirUtil;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

@Slf4j
@Component
public class AdminServiceImpl implements AdminService {

    @Autowired
    private ProjectRepo projectRepo;
    @Autowired
    private TissueTypeRepo tissueTypeRepo;
    @Autowired
    private CellTypeRepo cellTypeRepo;
    @Autowired
    private LesionTypeRepo lesionTypeRepo;
    @Autowired
    private PresetRepo presetRepo;
    @Autowired
    private ScoringRepo scoringRepo;
    @Autowired
    private CellRepo cellRepo;
    @Autowired
    private TelomereRepo telomereRepo;
    @Autowired
    private ImageDirUtil imageDirUtil;


    @Override
    public <T> T alterItem(T t) {
        Class clazz = t.getClass();
        AdminType adminType = AdminType.clazzToLookupType(clazz);
        switch (adminType){
            case PROJECT:
                return (T) projectRepo.save((Project) t);
            case TISSUE_TYPE:
                return (T) tissueTypeRepo.save((TissueType) t);
            case LESION_TYPE:
                return (T) lesionTypeRepo.save((LesionType) t);
            case CELL_TYPE:
                return (T) cellTypeRepo.save((CellType) t);
            case PRESET:
                return (T) presetRepo.save((Preset) t);
            default:
                throw new RuntimeException("bad lookupType: " + adminType);
        }
    }

    @Override
    public void deleteScoring(int scoringId) {
        log.info("Deleting scoring#{} and all references", scoringId);
        Scoring scoring = scoringRepo.findById(scoringId);
        Assert.notNull(scoring, "no scoring with id: " + scoringId);
        List<Cell> cells = cellRepo.findCellsByScoring(scoring);
        List<Telomere> telomeres = telomereRepo.findTelomeresByCellIn(new HashSet(cells));
        log.info("Deleting telomeres: {}", telomeres.stream().map(a -> a.getId()).collect(Collectors.toSet()));
        telomeres.forEach(t -> telomereRepo.delete(t));
        log.info("Deleting cells: {}", cells.stream().map(a -> a.getId()).collect(Collectors.toSet()));
        cells.forEach(c -> cellRepo.delete(c));
        log.info("Deleting scoringId#{}", scoring.getId());
        scoringRepo.delete(scoring);
        imageDirUtil.removeImages(scoring);
    }

    @Override
    public boolean scoringIdExists(int scoringId) {
        return scoringRepo.findById(scoringId) != null;
    }

    @Override
    public <T extends Named> List<T> findAllItems(Class<T> clazz) {
        return (List<T>) findAllItemsDelegate(clazz);
    }

    @Override
    public <T extends Named> Long idWithName(Class<T> clazz, String name) {
        List<T> existingItems = findAllItems(clazz);
        for (T candidate : existingItems){
            if (candidate.getName().trim().toLowerCase().equals(name.trim().toLowerCase())){
                long id = candidate.getId();
                return id;
            }
        }
        return null;
    }


    private<T> List findAllItemsDelegate(Class<T> clazz){
        AdminType adminType = AdminType.clazzToLookupType(clazz);
        switch (adminType){
            case PROJECT:
                return projectRepo.findAll();
            case TISSUE_TYPE:
                return tissueTypeRepo.findAll();
            case LESION_TYPE:
                return lesionTypeRepo.findAll();
            case CELL_TYPE:
                return cellTypeRepo.findAll();
            case PRESET:
                return presetRepo.findAll();
            default:
                throw new RuntimeException("bad lookupType: " + adminType);
        }
    }


    @Override
    public <T> boolean isReferenced(T t) {
        Class clazz = t.getClass();
        AdminType adminType = AdminType.clazzToLookupType(clazz);
        switch (adminType){
            case PROJECT:
                return !scoringRepo.findScoringsByProject((Project) t).isEmpty();
            case TISSUE_TYPE:
                return !cellRepo.findCellsByTissueType((TissueType) t).isEmpty();
            case LESION_TYPE:
                return !cellRepo.findCellsByLesionType((LesionType) t).isEmpty();
            case CELL_TYPE:
                return !cellRepo.findCellsByCellType((CellType) t).isEmpty();
            case PRESET:
                return !scoringRepo.findScoringsByPreset((Preset) t).isEmpty();
            default:
                throw new RuntimeException("bad lookupType: " + adminType);
        }
    }

    @Override
    public <T> void deleteItem(T t) {
        Assert.isTrue(!isReferenced(t), "Can't delete  -- is referenced: " + t);
        Class clazz = t.getClass();
        AdminType adminType = AdminType.clazzToLookupType(clazz);
        switch (adminType){
            case PROJECT:
                projectRepo.delete((Project) t);
                break;
            case TISSUE_TYPE:
                tissueTypeRepo.delete((TissueType) t);
                break;
            case LESION_TYPE:
                lesionTypeRepo.delete((LesionType) t);
                break;
            case CELL_TYPE:
                cellTypeRepo.delete((CellType) t);
                break;
            case PRESET:
                presetRepo.delete((Preset) t);
                break;
            default:
                throw new RuntimeException("bad lookupType: " + adminType);
        }
    }


}
