package edu.jhmi.telometer.view.query.table;

import edu.jhmi.telometer.bean.Cell;
import edu.jhmi.telometer.bean.ColumnSpec;
import edu.jhmi.telometer.bean.Project;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.bean.Telomere;
import edu.jhmi.telometer.service.api.TableService;
import edu.jhmi.telometer.util.CellTransientPopulator;
import edu.jhmi.telometer.util.DateUtil;
import edu.jhmi.telometer.util.DialogUtil;
import edu.jhmi.telometer.util.DoubleUtil;
import edu.jhmi.telometer.util.FileUtils;
import edu.jhmi.telometer.util.FrameUtil;
import edu.jhmi.telometer.util.IconUtil;
import lombok.extern.slf4j.Slf4j;
import org.springframework.util.Assert;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import java.awt.BorderLayout;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.prefs.Preferences;
import java.util.stream.Collectors;

@Slf4j
public class TablesPanel {

    private final JFrame frame = new JFrame("Export");
    private final Preferences preferences = Preferences.userRoot().node(getClass().getName());
    private static final String LAST_USER_TELOMETER_EXPORT_DIR = "LAST_USER_TELOMETER_EXPORT_DIR";
    private final JFileChooser exportFileChooser = new JFileChooser(preferences.get(LAST_USER_TELOMETER_EXPORT_DIR,
            new File(".").getAbsolutePath()));


    private GeneralTablePanel<Project> projectGeneralTablePanel = new GeneralTablePanel(Arrays.asList(PROJECT_COLUMN_SPECS));
    private GeneralTablePanel<Scoring> scoringGeneralTablePanel = new GeneralTablePanel(Arrays.asList(SCORING_COLUMN_SPECS));
    private GeneralTablePanel<Cell> cellGeneralTablePanel = new GeneralTablePanel(Arrays.asList(CELL_COLUMN_SPECS));
    private GeneralTablePanel<Telomere> telomereGeneralTablePanel = new GeneralTablePanel(Arrays.asList(TELOMERE_COLUMN_SPECS));
    private GeneralTablePanel[] generalTableCombos = {projectGeneralTablePanel, scoringGeneralTablePanel,
            cellGeneralTablePanel, telomereGeneralTablePanel};
    private JPanel mainPanel = new JPanel(new BorderLayout());

    private final JButton refreshProjectsButton = new JButton("Refresh", refreshIcon()); //projects
    private final JButton scoringsForwardButton = new JButton("Scorings", forwardIcon());

    private final JButton projectsBackwardButton = new JButton("Projects", backwardIcon()); //scorings
    private final JButton refreshScoringsButton = new JButton("Refresh", refreshIcon());
    private final JButton cellsForwardButton = new JButton("Cells", forwardIcon());

    private final JButton scoringsBackwardButton = new JButton("Scorings", backwardIcon()); //cells
    private final JButton telomeresForwardButton = new JButton("Telomeres", forwardIcon());

    private final JButton cellsBackwardButton = new JButton("Cells", backwardIcon()); //telomeres

    private final JButton exportProjectsButton = new JButton("Export", downloadIcon());
    private final JButton exportScoringsButton = new JButton("Export", downloadIcon());
    private final JButton exportCellsButton = new JButton("Export", downloadIcon());
    private final JButton exportTelomeresButton = new JButton("Export", downloadIcon());

    private final JTabbedPane tabbedPane = new JTabbedPane();
    private CellTransientPopulator cellTransientPopulator = new CellTransientPopulator();
    private TableService tableService;


    public TablesPanel() {
        FrameUtil.setInitialLocation(frame);
        exportFileChooser.setDialogTitle("Export");
        exportFileChooser.setAcceptAllFileFilterUsed(true);

        tabbedPane.addTab("Projects", projectsPanel());
        tabbedPane.addTab("Scorings", scoringsPanel());
        tabbedPane.addTab("Cells", cellsPanel());
        tabbedPane.addTab("Telomeres", telomeresPanel());

        mainPanel.add(tabbedPane, BorderLayout.CENTER);

        exportProjectsButton.addActionListener(l -> exportTableModel(projectGeneralTablePanel.getTableModel()));
        exportScoringsButton.addActionListener(l -> exportTableModel(scoringGeneralTablePanel.getTableModel()));
        exportCellsButton.addActionListener(l -> exportTableModel(cellGeneralTablePanel.getTableModel()));
        exportTelomeresButton.addActionListener(l -> exportTableModel(telomereGeneralTablePanel.getTableModel()));

        refreshProjectsButton.addActionListener(l -> projectGeneralTablePanel.setRows(tableService.findAllProjects()));
        refreshScoringsButton.addActionListener(l -> scoringGeneralTablePanel.setRows(tableService.findAllScorings()));

        scoringsForwardButton.addActionListener(l -> forwardButtonClicked(projectGeneralTablePanel, scoringGeneralTablePanel));
        cellsForwardButton.addActionListener(l -> forwardButtonClicked(scoringGeneralTablePanel, cellGeneralTablePanel));
        telomeresForwardButton.addActionListener(l -> forwardButtonClicked(cellGeneralTablePanel, telomereGeneralTablePanel));

        projectsBackwardButton.addActionListener(l -> backwardButtonClicked(scoringGeneralTablePanel, projectGeneralTablePanel));
        scoringsBackwardButton.addActionListener(l -> backwardButtonClicked(cellGeneralTablePanel, scoringGeneralTablePanel));
        cellsBackwardButton.addActionListener(l -> backwardButtonClicked(telomereGeneralTablePanel, cellGeneralTablePanel));

        refreshProjectsButton.setToolTipText("refresh projects -- show all");
        refreshScoringsButton.setToolTipText("refresh scorings -- show all");
        scoringsForwardButton.setToolTipText("view scorings for selected projects");
        cellsForwardButton.setToolTipText("view cells for selected scorings");
        telomeresForwardButton.setToolTipText("view telomeres for selected cells");
        exportProjectsButton.setToolTipText("export projects in table to a file");
        exportScoringsButton.setToolTipText("export scorings in table to a file");
        exportCellsButton.setToolTipText("export cells in table to a file");
        exportTelomeresButton.setToolTipText("export telomeres in table to a file");
        scoringsBackwardButton.setToolTipText("view the scorings for selected cells");
        cellsBackwardButton.setToolTipText("view the cells for the selected telomeres");
        projectsBackwardButton.setToolTipText("view the projects for the selected scorings");

        frame.add(getMainPanel());
        frame.pack();
    }

    private void exportTableModel(GeneralTableModel generalTableModel) {
        int response = exportFileChooser.showSaveDialog(mainPanel);
        if (response != JFileChooser.APPROVE_OPTION) {
            return;
        }
        File file = exportFileChooser.getSelectedFile();
        if (file.exists()){
            DialogUtil.displayInfoDialog(mainPanel, String.format("file already exists: %s", file.getAbsolutePath()));
            return;
        }
        preferences.put(LAST_USER_TELOMETER_EXPORT_DIR, file.getParent());
        FileUtils.writeListToFile(tableModelToListOfLists(generalTableModel), file);
    }

    private static List<List<String>> tableModelToListOfLists(GeneralTableModel tableModel){
        List<List<String>> lists = new ArrayList();
        List<String> header = new ArrayList();
        for (int column = 0; column < tableModel.getColumnCount(); column++){
            header.add(tableModel.getColumnName(column));
        }
        lists.add(header);
        for (int row = 0; row < tableModel.getRowCount(); row++){
            List<String> list = new ArrayList();
            lists.add(list);
            for (int column = 0; column < tableModel.getColumnCount(); column++){
                Object value = tableModel.getValueAt(row, column);
                value = value != null ? value : "";
                list.add(value.toString());
            }
        }
        return lists;
    }

    private <X, Y> void forwardButtonClicked(GeneralTablePanel<X> current, GeneralTablePanel<Y> next) {
        forwardOrBackwardButtonClicked(current, next, false);
    }

    private <X, Y> void backwardButtonClicked(GeneralTablePanel<X> current, GeneralTablePanel<Y> next) {
        forwardOrBackwardButtonClicked(current, next, true);
    }


    private <X, Y> void forwardOrBackwardButtonClicked(GeneralTablePanel<X> current, GeneralTablePanel<Y> next, boolean isBackward) {
        Set<X> selected = current.getSelectedObjects();
        if (selected.isEmpty()) {
            DialogUtil.displayInfoDialog(tabbedPane, "No rows were selected");
            return;
        }
        List<Y> nextRows = isBackward ? nextRowsFromSelectedBackward(selected) : nextRowsFromSelectedForward(selected);
        next.setRows(nextRows);
        int indexAddition = isBackward ? -1 : 1;
        tabbedPane.setSelectedIndex(tabbedPane.getSelectedIndex() + indexAddition);
    }

    private List nextRowsFromSelectedForward(Set selected) {
        Assert.isTrue(!selected.isEmpty(), "selection is empty");
        Class clazz = selected.iterator().next().getClass();
        if (clazz.equals(Project.class)) {
            return tableService.findScoringsByProjects((Set<Project>) (selected));
        } else if (clazz.equals(Scoring.class)) {
            return tableService.findCellsInScorings((Set<Scoring>) (selected));
        } else if (clazz.equals(Cell.class)) {
            return tableService.findTelomeresInCells((Set<Cell>) (selected));
        } else {
            throw new RuntimeException("class can't forward: " + clazz);
        }
    }

    private List nextRowsFromSelectedBackward(Set selected) {
        return new ArrayList(nextRowsFromSelectedBackwardDelegate(selected));
    }

    private Set nextRowsFromSelectedBackwardDelegate(Set selected) {
        Assert.isTrue(!selected.isEmpty(), "selection is empty");
        Class clazz = selected.iterator().next().getClass();
        if (clazz.equals(Scoring.class)) {
            return ((Set<Scoring>) selected).stream().map(s -> s.getProject()).collect(Collectors.toSet());
        } else if (clazz.equals(Cell.class)) {
            return ((Set<Cell>) selected).stream().map(c -> c.getScoring()).collect(Collectors.toSet());
        } else if (clazz.equals(Telomere.class)) {
            Set<Telomere> telomeres = (Set<Telomere>) selected;
            Set<Cell> cells = telomeres.stream().map(t -> t.getCell()).collect(Collectors.toSet());
            cellTransientPopulator.populateCellsTransientFields(cells, telomeres);
            return cells;
        } else {
            throw new RuntimeException("class can't backward: " + clazz);
        }
    }


    private JPanel projectsPanel() {
        JButton[] buttons = {refreshProjectsButton, exportProjectsButton, scoringsForwardButton};
        return tabbedPanel("Projects:", buttons, projectGeneralTablePanel);
    }

    private JPanel scoringsPanel() {
        JButton[] buttons = {projectsBackwardButton, refreshScoringsButton, exportScoringsButton, cellsForwardButton};
        return tabbedPanel("Scorings:", buttons, scoringGeneralTablePanel);
    }

    private JPanel cellsPanel() {
        JButton[] buttons = {scoringsBackwardButton, exportCellsButton, telomeresForwardButton};
        return tabbedPanel("Cells:", buttons, cellGeneralTablePanel);
    }


    private JPanel telomeresPanel() {
        JButton[] buttons = {cellsBackwardButton, exportTelomeresButton};
        return tabbedPanel("Telomeres:", buttons, telomereGeneralTablePanel);
    }

    private static JPanel tabbedPanel(String labelText, JButton[] buttons, GeneralTablePanel generalTablePanel) {
        JPanel buttonPanel = new JPanel();
        Arrays.stream(buttons).forEach(b -> buttonPanel.add(b));
        JLabel label = new JLabel(labelText);
        JPanel labelPanel = new JPanel(new BorderLayout());
        labelPanel.add(label, BorderLayout.WEST);
        JPanel northPanel = new JPanel(new BorderLayout());
        northPanel.add(labelPanel, BorderLayout.SOUTH);
        northPanel.add(buttonPanel, BorderLayout.CENTER);
        JPanel panel = new JPanel(new BorderLayout());
        panel.add(northPanel, BorderLayout.NORTH);
        panel.add(generalTablePanel.getMainPanel(), BorderLayout.CENTER);
        return panel;

    }


    public static ColumnSpec[] PROJECT_COLUMN_SPECS =
            {
                    ColumnSpec.<Project>builder()
                            .columnName("ProjectId")
                            .translator(p -> String.valueOf(p.getId()))
                            .build(),
                    ColumnSpec.<Project>builder()
                            .columnName("Name")
                            .translator(p -> p.getName())
                            .build(),
            };


    public static ColumnSpec[] SCORING_COLUMN_SPECS =
            {
                    ColumnSpec.<Scoring>builder()
                            .columnName("ScoringId")
                            .translator(s -> String.valueOf(s.getId()))
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("Name")
                            .translator(s -> s.getName())
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("ProjectId")
                            .translator(s -> String.valueOf(s.getProject().getId()))
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("ProjectName")
                            .translator(s -> s.getProject().getName())
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("Preset")
                            .translator(s -> s.getPreset().getName())
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("CreatedOn")
                            .translator(s -> DateUtil.formatDateTime(new Date(s.getCreatedOn())))
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("CyBackground")
                            .translator(s -> String.valueOf(s.getCyBackground()))
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("DapiBackground")
                            .translator(s -> String.valueOf(s.getDapiBackground()))
                            .build(),
                    ColumnSpec.<Scoring>builder()
                            .columnName("IsCy3Subtracted")
                            .translator(s -> String.valueOf(s.isCy3Subtracted()))
                            .build(),

            };


    public static ColumnSpec[] CELL_COLUMN_SPECS =
            {
                    ColumnSpec.<Cell>builder()
                            .columnName("CellId")
                            .translator(c -> String.valueOf(c.getId()))
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Name")
                            .translator(c -> c.getName())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("ScoringId")
                            .translator(c -> String.valueOf(c.getScoring().getId()))
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("ScoringName")
                            .translator(c -> c.getScoring().getName())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("X")
                            .translator(c -> c.getX() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Y")
                            .translator(c -> c.getY() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Width")
                            .translator(c -> c.getWidth() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Height")
                            .translator(c -> c.getHeight() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("LesionType")
                            .translator(c -> c.getLesionType().getName())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("CellType")
                            .translator(c -> c.getCellType().getName())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TissueType")
                            .translator(c -> c.getTissueType().getName())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Notes")
                            .translator(c -> c.getNotes())
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Sum")
                            .translator(c -> c.getSum() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Area")
                            .translator(c -> c.getArea() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Min")
                            .translator(c -> c.getMin() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Max")
                            .translator(c -> c.getMax() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("StdDev")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getStddev1000()) + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Mean")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getMean1000()) + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("Cy3Noise")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getCy3Noise1000()) + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("CellCount")
                            .translator(c -> c.getCellCount() != null ? c.getCellCount() + "" : "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereCount")
                            .translator(c -> c.getTelomereStat().getTelomereCount() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereSumSum")
                            .translator(c -> c.getTelomereStat().getTelomereSumSum() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereSumMin")
                            .translator(c -> c.getTelomereStat().getTelomereSumMin() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereSumMax")
                            .translator(c -> c.getTelomereStat().getTelomereSumMax() + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereSumMean")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getTelomereStat().getTelomereSumMean1000()) + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereAreaMean")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getTelomereStat().getTelomereAreaMean1000()) + "")
                            .build(),
                    ColumnSpec.<Cell>builder()
                            .columnName("TelomereSumSumToCellSumPercent")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getTelomereStat().getTelomereSumSumToCellSumPercent1000()) + "")
                            .build(),
            };



    public static ColumnSpec[] TELOMERE_COLUMN_SPECS =
            {
                    ColumnSpec.<Telomere>builder()
                            .columnName("TelomereId")
                            .translator(c -> String.valueOf(c.getId()))
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("TelemereNumber")
                            .translator(c -> c.getTelomereNumber() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("CellId")
                            .translator(c -> String.valueOf(c.getCell().getId()))
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("CellName")
                            .translator(c -> c.getCell().getName())
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("X")
                            .translator(c -> c.getX() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Y")
                            .translator(c -> c.getY() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Sum")
                            .translator(c -> c.getSum() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Area")
                            .translator(c -> c.getArea() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Min")
                            .translator(c -> c.getMin() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Max")
                            .translator(c -> c.getMax() + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("StdDev")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getStddev1000()) + "")
                            .build(),
                    ColumnSpec.<Telomere>builder()
                            .columnName("Mean")
                            .translator(c -> DoubleUtil.long1000ToDouble(c.getMean1000()) + "")
                            .build(),
            };


    private JPanel getMainPanel() {
        return mainPanel;
    }

    private static ImageIcon downloadIcon() {
        return IconUtil.getIcon("download");
    }

    private static ImageIcon refreshIcon() {
        return IconUtil.getIcon("refresh");
    }

    private static ImageIcon forwardIcon() {
        return IconUtil.getIcon("arrowForward");
    }

    private static ImageIcon backwardIcon() {
        return IconUtil.getIcon("arrowBackward");
    }

    public void setTableService(TableService tableService) {
        this.tableService = tableService;
    }

    public void setScoring(Scoring scoring) {
        List<Project> projects = Arrays.asList(scoring.getProject());
        List<Scoring> scorings = Arrays.asList(scoring);
        List<Cell> cells = tableService.findCellsInScorings(new HashSet(scorings));
        List<Telomere> telomeres = tableService.findTelomeresInCells(new HashSet(cells));
        projectGeneralTablePanel.setRows(projects);
        scoringGeneralTablePanel.setRows(scorings);
        cellGeneralTablePanel.setRows(cells);
        telomereGeneralTablePanel.setRows(telomeres);
        tabbedPane.setSelectedIndex(1);
    }


    public void display() {
        log.info("Displaying TablesPanel");
        frame.setVisible(true);
    }
}
