package edu.jhmi.telometer.view.query.tree;

import edu.jhmi.telometer.bean.Cell;
import edu.jhmi.telometer.bean.NodeType;
import edu.jhmi.telometer.bean.Project;
import edu.jhmi.telometer.bean.Scoring;
import edu.jhmi.telometer.bean.Telomere;
import edu.jhmi.telometer.bean.TreeRoot;
import edu.jhmi.telometer.interfce.Named;
import edu.jhmi.telometer.interfce.TreeObject;
import edu.jhmi.telometer.interfce.TreePanelListener;
import edu.jhmi.telometer.service.api.TreeService;
import edu.jhmi.telometer.thirdparty.SpringUtilities;
import edu.jhmi.telometer.util.DialogUtil;
import edu.jhmi.telometer.util.IconUtil;
import edu.jhmi.telometer.util.TreeGuiUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SpringLayout;
import javax.swing.SwingUtilities;
import javax.swing.event.MouseInputAdapter;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.MouseEvent;
import java.util.Enumeration;
import java.util.List;

public class TreeOnlyPanel {

    private static final Logger log = LoggerFactory.getLogger(TreeOnlyPanel.class);

    private TreeService treeService;

    private final TreeRoot ROOT_USER_OBJECT = new TreeRoot();
    private final Node root = new Node(ROOT_USER_OBJECT);
    private final DefaultTreeModel treeModel = new DefaultTreeModel(root);
    private final JTree tree = new JTree(treeModel);
    private final JPanel mainPanel = new JPanel(new BorderLayout());
    private final JButton refreshButton = new JButton("Refresh", IconUtil.getIcon("refresh"));

    private TreePanelListener treePanelListener;

    public void init()  {
        setToolTips();
        mainPanel.add(new JScrollPane(tree), BorderLayout.CENTER);
        mainPanel.add(refreshButton, BorderLayout.SOUTH);

        refreshButton.addActionListener(a -> refreshButtonClicked());

        tree.addTreeSelectionListener(new TreeSelectionListener() {
            @Override
            public void valueChanged(TreeSelectionEvent treeSelectionEvent) {
                if (treePanelListener == null){
                    return;
                }
                TreePath treePath = treeSelectionEvent.getPath();
                Node node = (Node) treePath.getLastPathComponent();
                TreeObject treeObject = node.getTreeObject();
                treePanelListener.treeObjectSelected(treeObject);
            }
        });

        final DefaultTreeCellRenderer cellRenderer = new DefaultTreeCellRenderer() {
            @Override
            public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus) {
                Component component = super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
                DefaultTreeCellRenderer label = (DefaultTreeCellRenderer) component;
                Node node = (Node) value;
                ImageIcon imageIcon = getIconForClass(node.getTreeObject().getClass());
                label.setIcon(imageIcon);
                return component;
            }
        };

        tree.setCellRenderer(cellRenderer);
        tree.addTreeSelectionListener(new TreeSelectionListener() {
            @Override
            public void valueChanged(TreeSelectionEvent treeSelectionEvent) {
                Node expanded = (Node) treeSelectionEvent.getPath().getLastPathComponent();
                TreeObject userObject = expanded.getTreeObject();
                addAllForUserObject(expanded, userObject);
            }
        });


    }

    private void setToolTips() {
        tree.setToolTipText("double click to expand the tree");
        refreshButton.setToolTipText("Refreshes the tree from the database");
    }

    private void refreshButtonClicked() {
        root.removeAllChildren();
        reload();
    }

    private void reload(){
        treeModel.reload();
    }

    private boolean addAllForUserObject(Node expanded, TreeObject treeObject) {
        Assert.notNull(expanded, "expanded is null");
        Assert.notNull(treeObject, "treeObject is null");
        Assert.notNull(treeService, "treeService is null");
        NodeType nodeType = TreeGuiUtil.nodeTypeForClass(treeObject.getClass());
        switch (nodeType){
            case ROOT:
                return addAll(expanded, treeService.findAllProjects());
            case PROJECT:
                return addAll(expanded, treeService.findScoringsByProject((Project) treeObject));
            case SCORING:
                return addAll(expanded, treeService.findCellsByScoring((Scoring) treeObject));
            case CELL:
                return addAll(expanded, treeService.findTelomeresByCell((Cell) treeObject));
            case TELOMERE:
                return false; //do nothing
            default:
                throw new RuntimeException("bad nodeType: " + nodeType);
        }
    }

    private boolean addAll(Node node, List<? extends TreeObject> list){
        if (node.getChildCount() > 0){
            return false;
        }
        for (TreeObject treeObject : list) {
            node.add(new Node(treeObject));
        }
        return true;
    }

    public void setScoring(Scoring scoring){
        refreshButtonClicked();
        tree.setSelectionPath(new TreePath(root));

        Project project = scoring.getProject();
        DefaultMutableTreeNode projectNode = findChildWithUserObject(root, project);
        Assert.notNull(projectNode, "no node for project: " + project);
        tree.setSelectionPath(new TreePath(treeModel.getPathToRoot(projectNode)));

        DefaultMutableTreeNode scoringNode = findChildWithUserObject(projectNode, scoring);
        Assert.notNull(scoringNode, "no node for scoring: " + scoringNode);
        tree.setSelectionPath(new TreePath(treeModel.getPathToRoot(scoringNode)));
    }


    private static DefaultMutableTreeNode findChildWithUserObject(DefaultMutableTreeNode parent, Object userObject) {
        Enumeration<TreeNode> enumeration = parent.children();
        while (enumeration.hasMoreElements()) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode) enumeration.nextElement();
            if (node.getUserObject().equals(userObject)) {
                return node;
            }
        }
        return null;

    }


    private static class Node extends DefaultMutableTreeNode {

        Node(TreeObject treeObject) {
            super(treeObject);
        }

        @Override
        public String toString() {
            TreeObject treeObject = getTreeObject();
            NodeType nodeType = TreeGuiUtil.nodeTypeForClass(treeObject.getClass());
            switch (nodeType){
                case TELOMERE:
                    return String.valueOf( ((Telomere) (treeObject)).getTelomereNumber());
                case PROJECT:
                case SCORING:
                case CELL:
                    return String.valueOf( ((Named) (treeObject)).getName());
                case ROOT:
                    return "Projects";
                default:
                    throw new RuntimeException("bad node type: " + nodeType);
            }
        }

        public TreeObject getTreeObject(){
            return (TreeObject) userObject;
        }
    }


    private ImageIcon getIconForClass(Class clazz) {
        return IconUtil.getIcon(getIconNameForClass(clazz));
    }

    private static String getIconNameForClass(Class clazz) {
        NodeType nodeType = TreeGuiUtil.nodeTypeForClass(clazz);
        return getIconNameForNodeType(nodeType);
    }

    private static String getIconNameForNodeType(NodeType nodeType) {
        switch (nodeType){
            case ROOT:
                return "projects";
            case PROJECT:
                return "project";
            case SCORING:
                return "scoring";
            case CELL:
                return "cell";
            case TELOMERE:
                return "telomere";
            default:
                throw new RuntimeException("Bad nodeType: " + nodeType);

        }
    }

    public void setTreeService(TreeService treeService) {
        this.treeService = treeService;
    }

    public JPanel getMainPanel(){
        return mainPanel;
    }

    public void setTreePanelListener(TreePanelListener treePanelListener) {
        this.treePanelListener = treePanelListener;
    }
}
